# -*- coding: utf8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Utils for class definition.
"""

from .module import get_calling_module_name
import sys


def _split_names(names):
  if isinstance(names, str):
    if ',' in names:
      names = [x.strip() for x in names.split(',')]
    else:
      names = names.split()
  return names


def comparable(properties, _stackdepth=0, decorate=None):
  """
  Creates a `__hash__()`, `__eq__()` and `__ne__()` method in the callers
  frame. The functions will hash/compare based on the specified
  *properties* (a whitespace/comma separate string or a list of attribute
  names).

  Optionally, the callers stackframe depth can be passed with the
  *_stackdepth* keyword-only argument. The keyword-only argument *decorate*
  may be used to decorate the generated functions.
  """

  properties = _split_names(properties)

  if decorate is None:
    def decorate(x):
      return x

  @decorate
  def __hash__(self):
    return hash(tuple(getattr(self, k) for k in properties))

  @decorate
  def __eq__(self, other):
    if type(self) != type(other):
      return False
    for k in properties:
      if getattr(self, k) != getattr(other, k):
        return False
    return True

  @decorate
  def __ne__(self, other):
    if type(self) != type(other):
      return True
    for k in properties:
      if getattr(self, k) != getattr(other, k):
        return True
    return False

  frame = sys._getframe(_stackdepth + 1)
  frame.f_locals['__hash__'] = __hash__
  frame.f_locals['__eq__'] = __eq__
  frame.f_locals['__ne__'] = __ne__


def repr(properties, _stackdepth=0, decorate=None):
  """
  Defines a `__repr__()` function in the callers frame that renders a
  string of the format `ClassName(attr1="value1", ...)`.
  """

  properties = _split_names(properties)
  parents = properties

  if decorate is None:
    def decorate(x):
      return x

  @decorate
  def __repr__(self):

    attrs = ', '.join('{}={!r}'.format(k, getattr(self, k)) for k in properties)
    return '{}({})'.format(type(self).__name__, attrs)

  frame = sys._getframe(_stackdepth + 1)
  frame.f_locals['__repr__'] = __repr__
  frame.f_locals['__repr_properties__'] = properties


def make_singleton(class_name, bool_value=True, module=None, base_class=object):
  """ Creates a new singleton object. """

  class SingletonType(object):
    __INSTANCE = None

    def __new__(cls):
      if cls.__INSTANCE is None:
        cls.__INSTANCE = object.__new__(cls)
      return cls.__INSTANCE

    def __repr__(self):
      return class_name

    def __bool__(self):
      return bool_value

    def __nonzero__(self):
      return bool_value

  SingletonType.__name__ = class_name
  SingletonType.__module__ = module or get_calling_module_name(1)
  SingletonType.__qualname__ = class_name
  return SingletonType()



# -*- coding: utf8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Provides tools for working with Python function objects, most prominently the
[[copy_function()]] function.
"""

from nr.collections import abc
import functools
import types
import sys
import traceback

__all__ = [
  'get_caller_name',
  'raise_kwargs',
  'except_format',
  'make_closure_cell',
  'new_closure',
  'copy_function'
]


def get_caller_name(_stackdepth=0):
  """
  Gets the name of the calling function.
  """

  return sys._getframe(_stackdepth + 1).f_code.co_name


def raise_kwargs(kwargs, name=None, _stackdepth=0):
  """
  Raises a [[TypeError]] indicating that the caller does not accept the
  specified keyword arguments. If *name* is `None`, it will be derived
  with [[get_caller_name()]].
  """

  if kwargs:
    if name is None:
      name = get_caller_name(_stackdepth + 1)
    key = next(iter(kwargs.keys()))

    raise TypeError('{!r} is an invalid keyword argument for {}()'
                    .format(key, name))


def except_format(func):
  """ Decorator for Exception `__str__()` methods. If an exception occurrs
  within the decorated function, the formatted traceback will be returned
  instead. """

  @functools.wraps(func)
  def wrapper(*args, **kwargs):
    try:
      return func(*args, **kwargs)
    except:
      return '<exception {}() failed> -- traceback below\n\n'.format(func.__name__) \
             + traceback.format_exc()

  return wrapper


def make_closure_cell(x):
  """
  Creates a new Python cell object that can usually be found in a functions
  `__closure__` tuple. The values in a `__closure__` must match the free
  variables defined in the function's code object's `co_freevars` member.

  x (any): The cell contents.
  return (cell): A function closure cell object containing *x*.
  """

  return (lambda: x).__closure__[0]


def new_closure(cell_values):
  """
  Creates a function closure from the specified list/iterable of value. The
  returned object is a tuple of cell objects created with #make_closure_cell().

  cell_values (iterable): An iterable of cell values.
  return (tuple of cell): A tuple containing only cell objects.
  """

  return tuple(map(make_closure_cell, cell_values))


def copy_function(
    function,
    code=None,
    globals=None,
    name=None,
    argdefs=None,
    closure=None):
  """
  Creates a new function object from the reference *function* where its
  members can be replaced using the specified arguments.

  # Closure

  If *closure* is a dictionary, only the free variables expected by the
  function are read from it. If a variable is not defined in the dictionary,
  its value will not be changed.

  If *closure* is a list/iterable instead it must have exactly as many
  elements as free variables required by the function's code object. If
  this condition is not satisfied, a #ValueError is raised.

  function (types.FunctionType): A function object.
  code (code): The functions new code object.
  globals (dict): Function global variables.
  name (str): Function name.
  argdefs (tuple): Function defaults.
  closure (closure): Function closure.
  """

  if not isinstance(function, types.FunctionType):
    raise TypeError('expected FunctionType, got {}'
                    .format(type(function).__name__))

  if code is None:
    code = function.__code__

  if globals is None:
    globals = function.__globals__

  if name is None:
    name = function.__name__

  if argdefs is None:
    argdefs = function.__defaults__

  if closure is None:
    closure = function.__closure__
  else:
    if isinstance(closure, abc.Mapping):
      closure = [
        closure.get(x, function.__closure__[i].cell_contents)
        for i, x in enumerate(function.__code__.co_freevars)]
    closure = new_closure(closure)
    if len(closure) != len(function.__code__.co_freevars):
      raise ValueError('function requires {} free closure, only '
                       '{} values are specified'.format(
                         len(function.__code__.co_freevars),
                         len(closure)
                       ))

  new_func = types.FunctionType(code, globals, name, argdefs, closure)
  functools.update_wrapper(new_func, function)
  return new_func



# -*- coding: utf8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Utilities for working with the Python [[typing]] module.
"""

from __future__ import absolute_import
import typing


def is_generic(
  x,                    # type: Any
  generic_types=None    # type: Union[None, typing._GenericAlias, Tuple[typing._GenericAlias]]
):
  # type: (...) -> bool
  """
  Checks if *x* is a specialized version of the specified *generic_types*.

  Examples:

  ```py
  assert is_generic(List[str])
  assert not is_generic(str)
  assert is_generic(List[str], List)
  assert is_generic(List[str], (Dict, List))
  assert is_generic(List, List)
  assert not is_generic(List, Dict)
  assert not is_generic(List[str], Dict)
  ```
  """

  has_special_form = hasattr(typing, '_SpecialForm')
  is_special_form = has_special_form and isinstance(x, typing._SpecialForm)

  if (not hasattr(x, '__origin__') or not hasattr(x, '__args__')) and \
      not is_special_form:
    return False
  if generic_types is None:
    return True
  if not isinstance(generic_types, (list, tuple)):
    generic_types = (generic_types,)

  def eq(a, b):
    # NOTE (@NiklasRosenstein): Due to a bug in Python 3.6, comparing
    #   typing.Union without args with another type causes a RecursionError.
    #   See https://bugs.python.org/issue29246
    if a is typing.Union:
      return b is typing.Union
    elif b is typing.Union:
      return False
    return a == b

  if is_special_form and generic_types:
    return x in generic_types

  for gtype in generic_types:
    # Special handling for Optional, subscripting it gives a Union.
    if gtype == typing.Optional and extract_optional(x) is not None:
      return True
    if eq(x, gtype) or eq(x.__origin__, gtype):
      return True
    if getattr(gtype, '__origin__', None) is not None and eq(x.__origin__, gtype.__origin__):
      return True
  return False


def get_generic_args(x):  # type: Any -> Tuple[Any]
  """
  Returns the arguments to a generic. If the generic is not specialized,
  the typevars are returned instead.
  """

  # Special handling for Optional, as subscripting it gives a Union with
  # two arguments.
  v = extract_optional(x)
  if v is not None:
    return (v,)

  return x.__args__ or x.__parameters__


def extract_optional(x):  # type: (Any, Any) -> Optional[Type]
  """
  Returns the type wrapped in [[Optional]] if *x* is an optional, otherwise
  returns `None`.
  """

  if getattr(x, '__origin__', None) is typing.Union:
    if len(x.__args__) == 2 and x.__args__[1] is type(None):
      return x.__args__[0]
  return None
